# from ._base_genetics._utils import *
# from ._base_genetics._validators import *
from ._base_genetics._base_genetic import Genetic
from ._base_ensemble_genetics import EnsembleGenetic


__all__ = ["Genetic", "EnsembleGenetic"]



import numpy as np

def validate_node_input(nodes):
    if type(nodes) != tuple:
        raise TypeError("fit method input must be a tuple of tuples")
    
    float_types = set([float, np.float, np.float32, np.float64])
    length = len(nodes)
    ids = []
    lats = []
    lons = []
    for node in nodes:
        if type(node) != tuple:
            raise TypeError("fit method input must be a tuple of tuples")
        if type(node[0]) != int:
            raise TypeError("fit method tuples first element must be integer")
        if type(node[1]) not in float_types:
            raise TypeError(f"fit method tuples second element must be a float representing the latitude {type(node[1])}") 
        if type(node[2]) not in float_types:
            raise TypeError("fit method tuples third element must be a float representing longitude")
        ids.append(node[0])
        lats.append(node[1])
        lons.append(node[2])
                
    if len(list(set(nodes))) != length:
        raise ValueError("Nodes can't be duplicated, there are nodes with same ids, latitude or longitude")



from itertools import combinations
import numpy as np
from ..._utils._utils import _cost
# from skroute._utils._utils import _cost
import random as rd

prueba = [1,2,3,4]

def create_combinations(rnd_sol_1, start_id):
    """
Create pair combinations from one route

Parameters
-----------
rnd_sol_1: list
    A list of ints with the Node's IDs

start_id_list: list
     A list with one integer. The ID representing the start Node
    """

    store_all_combinations = []
    rnd_sol_1 = rnd_sol_1[1:]
    route_index = range(len(rnd_sol_1)-1)
    list_of_n = list(combinations(route_index, 2))

    for swap1, swap2 in list_of_n:
        x_swap = rnd_sol_1[:]
        x_swap[swap1], x_swap[swap2] = x_swap[swap2], x_swap[swap1]
        store_all_combinations.append([start_id] + x_swap)
    
    return store_all_combinations


def cost_combinations(all_combinations, start_id, lenght, time_matrix, cost_matrix, max_time_work, extra_cost, people):
    """
Compute and order the cost of a list of routes

Parameters
-----------
all_combinations: list
    A list with differents routes

lenght: int
    The lenght that have all routes

time_matrix: dict
    it's a dict of dicts with the points to visit as a keys and
    the value is another dictionary with the points to visit and
    the value of going for the first point to the second. It's a
    dictionary representation of a matrix. For example if we
    have 3 ID 1, 2 and 3 the dict will be like this:

        {
        1:{
            1:0,
            2:x,
            3:y
            },
        2:{
            1:x,
            2:0,
            3:z
            },
        3:{
            1:y,
            2:z,
            3:0
            }
        }

    This is just a (3, 3) symmetric matrix with the cost in
    time from one point to another with column and index.

        1   2   3
    1   0   x   y
    2   x   0   z
    3   y   z   0

cost_matrix: dict
    For a extended explanation go up to time_matrix. it's
    the fuel cost to go from one point to another.

max_time_work: float32, default=8
    the number of ours that a employ can work per day. For example if it's 8 hours, the algorithm will
    force that a route have to finish after the 8 hours have been completed, making the employeed come
     back home. it's a time constraint.

extra_cost: float32, default=0
    if it's 0 anything happend. If it's >0 in combination with max_time_work when the max_time_work is 
    reached, extra_cost is applied. This add a extra cost to the solution each time that max_time_work is
    reached. It's like extra pay for the worker each time max_time_work is completed (journey).

people: int32, default=1
    The number of people that you use in each route, for example if you need two truck drivers. that's 
    another contstraint. That will multiply the time_costs and the extra_cost. Not the travel cost because
    it's assumed that both go in the same vehicle

    """
    cost_routes = []
    for route in all_combinations:
        route_copy = route[:]
        route_copy.append(start_id)
        cost_route = _cost(route_copy, lenght, time_matrix, cost_matrix, max_time_work, extra_cost, people)
        cost_routes.append(cost_route)

    cost_routes, routes = zip(*sorted(zip(cost_routes, all_combinations), key=lambda x: x[0]))
    return cost_routes, routes

def find_best_route(all_cost, all_routes):
    """
Find the best route of all routes.

Parameters
-----------

all_cost: list
    A list of integers with the cost of different routes

all_routes: list
    A list with the different routes, cost and route must have the same index position

Return
-------
A tuple with the best cost and the best route
    """
    cost_best_route = np.inf
    for i in range(len(all_cost)):
        if all_cost[i] < cost_best_route:
            cost_best_route = all_cost[i]
            best_route = all_routes[i]
    return cost_best_route, best_route


def random_swap_indices(length):
    """
Generate two random number between 0 and length

Parameters
length: int
    The highest random number that can be generated
    """
    rand_1 = int(rd.random()*length)+1
    rand_2 = int(rd.random()*length)+1
    while rand_1==rand_2:
        rand_2 = int(rd.random()*length)+1
    return rand_1, rand_2


def cross_route(mutated_route, length):
    """
Make a INPLACE reverse of the route passed

Parameters
-----------
mutated_route: list
    The route to mutate

lenght: int
    The lenght of the route to mutate

    """
    rand_1, rand_2 = random_swap_indices(length)

    if rand_1 > rand_2:
        mutated_route[rand_2:rand_1] = reversed(mutated_route[rand_2:rand_1])
    else:
        mutated_route[rand_1:rand_2] = reversed(mutated_route[rand_1:rand_2])


def var_lower_length(tabu_lenght, tabu_var):
    """
Validation function that assert that tabu_var isn't higher number than tabu_lenght

Parameters:
-----------
tabu_lenght: int
tabu_var: int
    """
    
    if tabu_var <= tabu_lenght:
        return tabu_var
    else:
        raise ValueError("tabu_var can't be higher than tabu_lenght")



"""
Simulated Annealing class

Notes
-----
In the future we can improve the speed with Cython

- Its not returning sometimes the rute accordingly to the cost.


Authors
-------
2020: Alberto Rubiales <al.rubiales.b@gmail.com>
"""

from ._utils_nrbs import _compute_mean_std_dicts, _priority_function, _priority_connection, Node, _check_route
import os, sys
sys.path.append("/".join(os.path.abspath(__file__).split("/")[:-3]))
from ..._validators._validators import (_float_validator, _intenger_validator, _validate_route_example, _validate_dict_of_dicts)
from ..._utils._utils import _cost
from copy import deepcopy
import numpy as np
#Node Ranking Based on Stats
class NRBS:
    """
An Heuristic Algorithm, Node Ranking Based on Stats (NRBS) Class

Parameters:
------------

mean_priority: float
    Strongly recommended a number between 0 and 2, The ponderation of the mean for 
    all points, higher number make points with higher mean (more distance from
    others) have large values and priority selection

std_priority: float
    Strongly recommended a number between 0 and 2, The ponderation of the standar
    deviation for all points, higher number make points with higher deviation
    (more diferents of distance between points) have larger values and priority
    selection

mean_connection: float
    Strongly recommended a number between 0 and 2, The ponderation of the mean for 
    all points, higher number make points with higher mean (more distance from
    others) have large values and priority to connect

std_connection: float
    Strongly recommended a number between 0 and 2, The ponderation of the standar
    deviation for all points, higher number make points with higher deviation
    (more diferents of distance between points) have larger values and priority
    to connect

distance_weight: float
    Strongly recommended a number between 0 and 2, The ponderation of the distance 
    between all points, higher distance_weight make points with higher distances
    between them have large values, this reduce the influce of the mean_connection
    and the std_connections parameters depending of the distance.


---------------------------------------------------------------------------------------
Methods

fit(start_point_id, ids_route, cost_matrix)
    train the algorithm and give back tghe best solution

    Parameters
    -----------
    start_point_id: int
        The point of start
    
    ids_route: list
        A list with all points

    cost_matrix: dict
        it's a dict of dicts with the points to visit as a keys and
        the value is another dictionary with the points to visit and
        the value of going for the first point to the second. It's a
        dictionary representation of a matrix. For example if we
        have 3 ID 1, 2 and 3 the dict will be like this:

            {
            1:{
                1:0,
                2:x,
                3:y
                },
            2:{
                1:x,
                2:0,
                3:z
                },
            3:{
                1:y,
                2:z,
                3:0
                }
            }

        This is just a (3, 3) symmetric matrix with the cost in
        time from one point to another with column and index.

            1   2   3
        1   0   x   y
        2   x   0   z
        3   y   z   0

    Returns
    --------
    A very optimized route that solve the TSP/VRP problem.
    """

    def __init__(self, mean_priority, std_priority, mean_connection, std_connection, distance_weigth):
        self._mean_priority = _float_validator(mean_priority, "mean_priority")
        self._std_priority = _float_validator(std_priority, "std_priority") 
        self._mean_connection = _float_validator(mean_connection, "mean_Connection")
        self._std_connection = _float_validator(std_connection, "std_connection")
        self._distance_weigth = _float_validator(distance_weigth, "distance_weight")

    def fit(self, start_node_id, ids_node, cost_matrix):
        _intenger_validator(start_node_id, "start_point_id")
        _validate_dict_of_dicts(cost_matrix, "cost_matrix")
        _validate_route_example(ids_node, cost_matrix, cost_matrix)
        _lenght_route = len(ids_node)

        mean_dict, std_dict = _compute_mean_std_dicts(ids_node, cost_matrix, _lenght_route)
        priority_routes = _priority_function(ids_node, mean_dict, std_dict, self._mean_priority, self._std_priority)
        priority_routes_conections = _priority_connection(priority_routes, cost_matrix, mean_dict, std_dict, self._mean_connection, self._std_connection, self._distance_weigth)

        #Cogemos y agrupamos cada ruta con su conexión mejor y tenemos un counter_dict que cuenta cuando están conectadas y cuando no
        node_values = {i:Node(i) for i in ids_node}

        for i in range(2):
            for k, _ in priority_routes.items():
                if len(node_values[k]) < 2:
                    temp_connection = priority_routes_conections[k]
                    temp_keys = list(temp_connection.keys())
                    sol_found = False
                    counter = 0
                    while counter < _lenght_route-1 and not sol_found:
                        to_connect = temp_keys[counter]
                        temp_node_values = deepcopy(node_values)
                        if len(node_values[to_connect]) < 2:
                            temp_node_values[k].neighbours = to_connect
                            temp_node_values[to_connect].neighbours = k
                            if to_connect not in node_values[k].neighbours and _check_route(temp_node_values, _lenght_route):
                                node_values[k].neighbours = to_connect
                                node_values[to_connect].neighbours = k
                                sol_found = True
                        counter += 1
        result = _check_route(node_values, 19)
        cost = _cost(result, _lenght_route, cost_matrix, cost_matrix, np.inf, 0, 1)
        result.append(start_node_id)
        return cost, result
        






import sys
sys.path.append("/home/rubiales/Desktop/Projects/scikit-route/skroute/metaheuristics/tabu_search")
from _tabu_search import TabuSearch
from skroute.datasets import load_barcelona, load_madrid, load_valencia, load_alicante_murcia, load_finland
from skroute.preprocessing import dfcolumn_to_dict
from _utils_tabu import *

df_barcelona = load_barcelona()["DataFrame"]
matrix_cost = dfcolumn_to_dict(df_barcelona, "id_origin", "id_destinity", "cost")
matrix_time = dfcolumn_to_dict(df_barcelona, "id_origin", "id_destinity", "hours")

route_exampl = list(dict.fromkeys(matrix_cost).keys())


ts = TabuSearch()
ts.fit(route_exampl, matrix_time, matrix_cost)

combs = create_combinations(route_exampl, [route_exampl[0]])

cost_combs = cost_combinations(combs, 19, matrix_time, matrix_cost, 8, 10, 1)




from ._base_brute_force._base_brute_force import BruteForce

#NOTE: C Version looks like Python version wich is normal, because both are pretty much the same.
#NOTE: We can make a brute force with clusters maybe works

__all__ = ["BruteForce"]



import tensorflow as tf
import numpy as np


def generate_weights(size):
    """
Create random weights for the neural network

Parameters:
------------
size int
    The number of weights of the neural network

Return:
-------
A two dimensional tensor of weights
    """
    return tf.random.uniform(shape=[size, 2])

def get_neighbor(center, radix, domain):
    """
Based on the center, the radix and the domain to find best neighbourhood and
help update the weights in the future.

Parameters:
------------

center int
    The id of the node to find the neighbors

radix int
    The radiux of search neighbors

domain array
    The weights that will be used to find networks

Return:
--------
A gaussian distribution that will be used to updates the weights on the future

    """

    if radix < 1:
        radix = 1

    deltas = tf.abs(center - tf.range(domain))
    distances = tf.minimum(deltas, domain - deltas)

    return tf.exp(-(distances * distances) / (2* (radix * radix)))


def euclidean_distance(weights, node):
    """
Based on weights compute the euclidean distance between nodes

Parameters:
------------

weights array
    The weights used to compute the distance between nodes

nodes tuple
    A node with the latitude and the longitude to compute the distance
 
 Return:
 -------
 The distance between a node and the rest
    """
    return tf.linalg.norm(weights - node, axis=1)

def get_closest(weights, node):
    """
Based on weights returns the closest neigbour to the node

Parameters:
------------

weights array
    The weights used to compute the distance between nodes

nodes tuple
    A node with the latitude and the longitude to compute the distance
 
 Return:
 -------
 The position index of the best neighbor of node
    """
    return tf.cast(tf.math.argmin(euclidean_distance(weights, node)), tf.int32)

def random_gen(length):
    """
    Generate a random number between 0 and length
    """
    return int(tf.random.uniform(shape=[1])*length)




from ._utils_som import *
from ._som_validators import * 
import tensorflow as tf
from ..._validators._validators import _intenger_validator, _zero_one_validator
from tqdm import tqdm



class SOM:
    """
SOM (Self Organizing Map) class is a Neuronal Network approach to the TSP/VRP problem

Parameters:
------------
    
units: int, default=None
    The neurons of our Neural Network, if it's None the units will be the length of the route
    multiplyed by eight

radius: int, default=None
    The area of search to obtain new neighbours, if it's None the radius will be the length of
    the route multiplyed by eight

radius_decay: float, default=0.9991
    The decay of the radius. In order to find the most optimal solution, at the end is good
    to only look the most near neighbours.

lr: float, default=0.8
    The learning rate, is how aggressive the update of the weight is, higher learning rate, more
    aggressive is the update of the weights

lr_decay: float, default=0.9991
    The decay of the learning rate. In order to find the most optimal solution, at the end is good
    to have a low learning rate.


--------------------------------------------------------------------------

Methods:
--------

fit(nodes, epochs)
    Execute the algorithm and give back the best route find it

nodes: tuple
    A tuple of tuples, each tuple is a Node with the first element the ID, the sencond the latitude
    and the third then longitude. For example if we have a route with three points, the tuple will
    be like this:

    nodes = (
        (1, 0.459887, 14.345767),
        (2, 0.634534, 12.575462),
        (3, 0.256765, 9.734435),
    )


epochs: int, default=10_000
    The times that the Neural network will update the weights trying to find the optimal solution


Returns
--------
A very optimized route that solve the TSP/VRP problem.


Example
--------
from sklearn_route.datasets import load_barcelona
from sklearn_route.preprocessing import normalize, df_to_tuple
from sklearn_route.metaheuristics.som import SOM

df_barcelona = load_barcelona()["DataFrame"]
df_barcelona = df_barcelona[["id_origin", "lat_origin", "lon_origin"]].drop_duplicates()
df_barcelona = normalize(df_barcelona, "lat_origin", "lon_origin")

route = df_to_tuple(df_Barcelona)

som = SOM()
result = som.fit(route)

#Printing the best route
print(result)
    """

    def __init__(self, units=None, radius=None, radius_decay=0.9991, lr=0.8, lr_decay=0.9991):
        self.units = None if not units else _intenger_validator(units, "units")
        self.radius = None if not radius else _intenger_validator(radius, "radius")
        self.radius_decay= _zero_one_validator(radius_decay, "radius_decay")
        self.lr = _zero_one_validator(lr, "lr")
        self.lr_decay = _zero_one_validator(lr_decay, "lr_decay")


    def fit(self, nodes, epochs=10_000):
        epochs = _intenger_validator(epochs, "epochs")
        validate_node_input(nodes)
        length = len(nodes)
        if self.units is None:
            self.units = 8 * length

        if self.radius is None:
            self.radius = length * 8
        
        weights = generate_weights(self.units)


        for i in tqdm(range(epochs)):
            node = nodes[random_gen(length)][1:]
            node_idx = get_closest(weights, node)
            gaussian = get_neighbor(node_idx, int(self.radius//10), weights.shape[0])
            weights += tf.expand_dims(tf.cast(gaussian, tf.float32), axis=1) * self.lr * (node - weights)

            self.lr *= self.lr_decay
            self.radius *= self.radius_decay
            if self.radius < 1:
                print('\nRadius has completely decayed, finishing execution',
                'at {} iterations'.format(i))
                break
            if self.lr < 0.001:
                print('\nLearning rate has completely decayed, finishing execution',
                'at {} iterations'.format(i))
                break
        route = []
        for node in nodes:
            route.append(get_closest(weights, node[1:]))

        route = [int(i) for i in list(list(zip(*list(zip(*sorted(zip(route, nodes))))[1]))[0])]

        return route




def _zero_one_validator(x, function):
    message = function + " must be a float between zero and one"
    if x < 0.0 or x > 1.0:
        raise ValueError(message)
    return x


def _intenger_validator(x, function):
    message = function + " must be positive intenger"
    if isinstance(x, int) and x > 0:
        return x
    raise TypeError(message)


def _early_stopping_validator(x):
    if x:
        if x <=0 or not isinstance(x, int):
            raise ValueError("Early stopping must be a positive number bigger than 0 or None for disable it")
        return x
    return x


def _float_validator(x, function):
    message = function + " must be a positive float"
    if x < 0 or type(x) != float:
        raise ValueError(message)
    return x


def _validate_dict_of_dicts(x, function):
    message = function + " must be a dict of dict representing a simmetric matrix"

    if not isinstance(x, dict) and not isinstance(x[list(x.keys())], dict):
        raise TypeError(message)

    for k in x.keys():
        if x[k].keys() != x.keys():
            raise ValueError(message)

def _validate_route_example(x, fuel_costs, time_costs):
    for i in x:
        if not isinstance(i, int):
            raise TypeError("The route provided must be a list of integers")

    x_sorted = sorted(x)
    dict_sorted = sorted(list(fuel_costs.keys()))

    if x_sorted != dict_sorted:
        raise ValueError("The ints in the random route provided must be in the dicts as keys")

    if len(x) == len(list(fuel_costs.keys())) == len(list(time_costs.keys())):
        pass
    else:
        raise ValueError("The dicts must have the same dimension and the route example must have the same lenght as the dicts keys")




from ._base_simulated_annealing._base_simulated_annealing import SimulatedAnnealing
from multiprocessing import Pool
from ..._validators._validators import _intenger_validator

class EnsembleSimulatedAnnealing(SimulatedAnnealing):
    """
The Metaheuristic algorithm, Simmulated Annealing class.

Parameters:
-----------

temp: float32, default=12
    The temperature parameter. In the algorithm the temperature will 
    decrease till reach the tol parameter to give you the best
    solution.

neighbours: int32 , default=250
    Given a route the exchanges that will make in at the same temperature.
    You can think as the neighbours that will visit at a particular
    temperature.

delta: float32, default=0.78
    A number between 0 and 1 the highest the number more slow the
    the temperature decrease, and more solution the algorithm try.
    If it's close to 0 the temperature decrease fast and the
    algorithm converge faster but look for less optimal solutions

tol: float32, default=1.29
    The tolerance, the minimun temperature allow by the algorithm
    one time the "temp" is equal to "tol" the algorithm finish.

max_time_work: float32, default=8
    the number of ours that a employ can work per day. For example
    if it's 8 hours, the algorithm will force that a route have to
    finish after the 8 hours have been completed, making the
    employeed come back home. it's a time constraint.

extra_cost: float32, default=0
    if it's 0 anything happend. If it's >0 in combination with
    max_time_work when the max_time_work is reached, extra_cost is
    applied. This add a extra cost to the solution each time that
    max_time_work is reached. It's like extra pay for the worker 
    each time max_time_work is completed (journey).

people: int32, default=1
    The number of people that you use in each route, for example if
    you need two truck drivers. that's another contstraint. That
    will multiply the time_costs and the extra_cost. Not the travel
    cost because it's assumed that both go in the same vehicle.

--------------------------------------------------------------------------

Methods:
--------

fit(self, list route_example, dict time_matrix, dict fuel_matrix)
    Execute the algorithm and give back the best solution

    Parameters
    ----------
    route_example: list
        it's a list that contain a random example of ints ids routes. But it's
        mandatory that the origin place, be the first id of the list

    time_costs: dict
        it's a dict of dicts with the points to visit as a keys and
        the value is another dictionary with the points to visit and
        the value of going for the first point to the second. It's a
        dictionary representation of a matrix. For example if we
        have 3 ID 1, 2 and 3 the dict will be like this:

            {
            1:{
                1:0,
                2:x,
                3:y
                },
            2:{
                1:x,
                2:0,
                3:z
                },
            3:{
                1:y,
                2:z,
                3:0
                }
            }

        This is just a (3, 3) symmetric matrix with the cost in
        time from one point to another with column and index.

            1   2   3
        1   0   x   y
        2   x   0   z
        3   y   z   0
    
    fuel_costs: dict
        For a extended explanation go up to time_matrix. it's
        the fuel cost to go from one point to another.

    Returns:
    --------
    A very optimized route that solve the MTSPTW problem.


---------------------------------------------------------------------------

Attributes
----------

history_: The best solution in each iteration of the
          generation or the decrease of the loss.

Example:
--------
from sklearn_route.datasets import load_barcelona
from sklearn_route.preprocessing import matrix_to_dict
from sklearn_route.metaheuristics.simmulated_annealing import EnsembleSimulatedAnnealing

df_barcelona = load_barcelona()

#Dataset - id of origin - id of destiny - column to transform (in this case the hour)
time_matrix = matrix_to_dict(df_barcelona, "id_origin", "id_destinity", "hora")

#Dataset - id of origin - id of destiny - column to transform (in this case the cost)
cost_matrix = matrix_to_dict(df_barcelona, "id_origin", "id_destinity", "hora")

#Create a random route, it's will needed to initiate the algorithm
route_example = list(dict.fromkeys(cost_matrix_df).keys())

#Instantiate the algorithm
esa = EnsembleSimulatedAnnealing()

#random route - time_matrix - cost_matrix
result = esa.fit(route_example, time_matrix, cost_matrix)

#Printing the best route
print(result)

#Printing the loss function
print(esa.history_)
    """

    def __init__(self, n_simulateds=20,  temp:float=12.0, neighbours:int=250, delta:float=0.78, tol:float=1.29,
                max_time_work:float=8.0, extra_cost:float=10.0, people:int=1, n_jobs:int=1):
        super().__init__(temp, neighbours, delta, tol, max_time_work, extra_cost, people)
        self.n_simulateds = _intenger_validator(n_simulateds, "n_simulateds")
        self.n_jobs = _intenger_validator(n_jobs, "n_jobs")
    
    def fit(self, route_example:list, time_matrix:dict, cost_matrix:dict) -> list:
        if self.n_jobs > 1:
            pool = Pool(processes=self.n_jobs)
            list_to_pool = []
            for i in range(self.n_simulateds):
                list_to_pool.append([SimulatedAnnealing(self.temp, self.neighbours, self.delta, self.tol,
                                                        self.max_time_work, self.extra_cost, self.people),
                                    route_example, time_matrix, cost_matrix])
            list_results = pool.starmap(SimulatedAnnealing.fit, list_to_pool)
            min_value = list_results[0][0]
            result = list_results[0]
            for i in range(self.n_simulateds):
                if list_results[i][0] < min_value:
                    result = list_results[i]
            return result
        else:
            result = []
            value = []
            for i in range(self.n_simulateds):
                algo_result = super().fit(route_example, time_matrix, cost_matrix)
                result.append(algo_result)
                value.append(algo_result[0])
            return result[value.index(min(value))]




# import sys
# sys.path.append("/home/rubiales/Desktop/Projects/scikit-route/skroute/_validators")

# from _validators import (_intenger_validator,_zero_one_validator, _float_validator,
#                                         _validate_dict_of_dicts, _validate_route_example)


from ..._validators._validators import (_intenger_validator,_zero_one_validator, _float_validator,
                                        _validate_dict_of_dicts, _validate_route_example)
from ..simulated_annealing._base_simulated_annealing._utils_sa import _swap_route

import random as rd
from collections import deque
from ._utils_tabu import *
from ..._utils._utils import _final_cost

# from _utils_tabu import *


# def _swap_route(route, r1, r2):

#     temporal = 0
#     temporal = route[r1]
#     route[r1] = route[r2]
#     route[r2] = temporal
#     return route




class TabuSearch():
    """
Instantaite the Tabu Search class.

Tabu Search is a metaheuristic algorithm that solve TSPTW/VRPTW 

Parameters:
------------
searchs: int, default=1250
    The number of neighbourhood route searchs, more searchs will lead to a better result but with an 
    associate time computation cost.

p_m : float, default=0.6
    it's mutate, random probabilies are choosen when it will take place. If the random numbers are 2
    and 8 for example, the numbers located at that index will swap positions.

tabu_length: int, default=45
    it's the lenght of tabu list where routes are stored in order to avoid visits them again.

tabu_var: int, default=10
    Variations in the lenght of tabu_length, can't be higher number than tabu_length

max_time_work: float32, default=8
    the number of ours that a employ can work per day. For example if it's 8 hours, the algorithm will
    force that a route have to finish after the 8 hours have been completed, making the employeed come
     back home. it's a time constraint.

extra_cost: float32, default=0
    if it's 0 anything happend. If it's >0 in combination with max_time_work when the max_time_work is 
    reached, extra_cost is applied. This add a extra cost to the solution each time that max_time_work is
    reached. It's like extra pay for the worker each time max_time_work is completed (journey).

people: int32, default=1
    The number of people that you use in each route, for example if you need two truck drivers. that's 
    another contstraint. That will multiply the time_costs and the extra_cost. Not the travel cost because
    it's assumed that both go in the same vehicle

ts = TabuSearch(searchs= 1250, p_m = 0.6, tabu_length=45, tabu_var=10, max_time_work=6, extra_cost=12.83)

--------------------------------------------------------------------------------------------------------------

Methods:
--------

fit(self, list random_route, dict time_matrix, dict cost_matrix)
    Execute the algorithm and give back the best solution

    Parameters
    ----------
    random_route: list
        it's a list that contain a random example of ints ids routes. But it's
        mandatory that the origin place, be the first id of the list

    time_matrix: dict
        it's a dict of dicts with the points to visit as a keys and
        the value is another dictionary with the points to visit and
        the value of going for the first point to the second. It's a
        dictionary representation of a matrix. For example if we
        have 3 ID 1, 2 and 3 the dict will be like this:

            {
            1:{
                1:0,
                2:x,
                3:y
                },
            2:{
                1:x,
                2:0,
                3:z
                },
            3:{
                1:y,
                2:z,
                3:0
                }
            }

        This is just a (3, 3) symmetric matrix with the cost in
        time from one point to another with column and index.

            1   2   3
        1   0   x   y
        2   x   0   z
        3   y   z   0
    
    cost_matrix: dict
        For a extended explanation go up to time_matrix. it's
        the fuel cost to go from one point to another.

    Returns:
    --------
    A very optimized route that solve the TSPTW/VRPTW problem.

--------------------------------------------------------------------------------------------------------------

Attributes
----------

history_: The best solution in each iteration of the searchs or the decrease of the loss.

Example
--------
from sklearn_route.datasets import load_barcelona
from sklearn_route.preprocessing import matrix_to_dict
from sklearn_route.metaheuristic.tabu_search import TabuSearch

df_barcelona = load_barcelona()

#Dataset - id of origin - id of destiny - column to transform (in this case the hour)
time_matrix = matrix_to_dict(df_barcelona, "id_origin", "id_destinity", "hora")

#Dataset - id of origin - id of destiny - column to transform (in this case the cost)
cost_matrix = matrix_to_dict(df_barcelona, "id_origin", "id_destinity", "hora")

#Create a random route, it's will needed to initiate the algorithm
route_example = list(dict.fromkeys(cost_matrix_df).keys())

#Instantiate the algorithm
ts = TabuSearch(searchs= 1250, p_m = 0.6, tabu_length=45, tabu_var=10, max_time_work=6, extra_cost=12.83)

#random route - time_matrix - cost_matrix
result = ts.fit(route_example, time_matrix, cost_matrix)

#Printing the best route
print(result)

#Printing the loss function
print(ts.history_)
    """


    def __init__(self, searchs=1250, p_m=0.6, tabu_length=45, tabu_var=10, max_time_work=8., people=1, extra_cost=1.):
        self.searchs = _intenger_validator(searchs, u"searchs")
        self.p_m = _zero_one_validator(p_m, u"p_m")
        self.tabu_length = _intenger_validator(tabu_length, u"lenght_tabu")
        self.tabu_var = var_lower_length(tabu_length, _intenger_validator(tabu_var, u"tabu_var"))
        self.max_time_work = _float_validator(max_time_work, u"max_time_work")
        self.people = _intenger_validator(people,  u'people')
        self.extra_cost = _float_validator(extra_cost, u"extra_cost")

    def fit(self, random_route, time_matrix, cost_matrix):
        _validate_dict_of_dicts(time_matrix, u"time_matrix")
        _validate_dict_of_dicts(cost_matrix, u"cost_matrix")
        _validate_route_example(random_route, time_matrix, cost_matrix)

        start_id = random_route[0]
        length = len(random_route)
        length_less1 = length - 1
        all_solutions_route = []
        all_solutions_cost = []
        tabu_costs = deque()
        tabu_routes = deque()
        low_tabu = self.tabu_length - self.tabu_var
        high_tabu = self.tabu_length + self.tabu_var
        rnd_sol_1 = random_route[:]
        

        for i in range(1, self.searchs):
            all_routes = create_combinations(rnd_sol_1, start_id)

            cost_routes, all_routes = cost_combinations(all_routes, start_id, length, time_matrix, cost_matrix, self.max_time_work, self.people, self.extra_cost) #TODO: hay que cambiar por self
            best_cost_it = cost_routes[0]
            best_route_it = all_routes[0]

            if all_solutions_cost:
                best_cost, best_route = find_best_route(all_solutions_cost, all_solutions_route) 
                t = 0
                while best_cost_it in tabu_costs and t < length_less1:
                    if best_cost < best_cost_it:
                        best_cost_it = best_cost
                        best_route_it = best_route[:]
                        break
                    else:
                        best_cost_it = cost_routes[t]
                        best_route_it = all_routes[t]
                        t +=1
                
            
            if len(tabu_costs) >= self.tabu_length:
                if best_cost_it not in tabu_costs:
                    while len(tabu_costs) >= self.tabu_length:
                        tabu_costs.popleft()
                        tabu_routes.popleft()
                tabu_costs.append(best_cost_it)
                tabu_routes.append(best_route_it)
            else:
                tabu_routes.append(best_route_it)
                tabu_costs.append(best_cost_it)
            
            all_solutions_cost.append(best_cost_it)
            all_solutions_route.append(best_route_it)

            mod_iterations_1 = i%20  
            mod_iterations_2 = i%40 
            mod_iterations_3 = i%30 

            if mod_iterations_3 == 0 and best_cost_it > best_cost: 
                best_cost_it = best_cost
                best_route_it = best_route[:]

            if rd.random() <= self.p_m:
                best_route_it = best_route_it[:]
                cross_route(best_route_it, length)
                rnd_sol_1 = best_route_it
            else:
                rnd_sol_1 = best_route_it[:]

            
            if mod_iterations_2 == 0:
                rand_1, rand_2 = random_swap_indices(length_less1)
                rand_1_1, rand_3_1 = random_swap_indices(length_less1)
                rand_3, rand_4 = random_swap_indices(length_less1)
                
                best_route = _swap_route(best_route[:], rand_1, rand_2)
                best_route = _swap_route(best_route[:], rand_1_1, rand_2)
                best_route = _swap_route(best_route[:], rand_3, rand_4)
                best_route = _swap_route(best_route[:], rand_3_1, rand_4)

                rnd_sol_1 = best_route[:]

            
            if mod_iterations_1 == 0:
                self.tabu_length = rd.randint(low_tabu, high_tabu)

        
        final_cost, final_route = find_best_route(all_solutions_cost, all_solutions_route)
        final_route = _final_cost(final_route, length, time_matrix, self.max_time_work)
        self.history_ = list(zip(all_solutions_cost, all_solutions_route))
        return final_cost, final_route





def _std(var:list, mean:float, length:int):
    """
Compute the Standar Deviation of a point

Parameters:
------------

var: list
    The variance between all points

mean: float
    The mean of the point

lenght: int
    The lenght of all routes

Return
-------
Return the standar deviation of a point
    """

    summatory = 0
    for i in range(length):
        summatory += (var[i] - mean)**2

    return (summatory / length)**0.5

def _priority_function(random_route:list, mean_routes:dict, std:dict, mean_weight:float, std_weight:float)-> dict:
    """
Give us the order of  priority routes to select to look for connections

Parameters
-----------

random_route: list
    The list of all points to connect

mean_routes: dict
    A dict with the mean distance/cost of all points one each other

std_routes: dict
    A dict with the standar deviation distance/cost of all points one each other

mean_weight: float
    The ponderation of the mean for all points, higher number make points
    with higher mean (more distance from others) have large values and priorice
    their connections

std_weight: float
    The ponderation of the standar deviation for all points, higher number
    make points with higher deviation (more diferents of distance between points)
    have larger values and priorice their connections

Return
-------
A Dict of oredered by priority
    """

    priorities = {}
    for id_route in random_route:
        priorities[id_route] = (mean_routes[id_route]**mean_weight) * (std[id_route]**std_weight)

    ordered_prior_route = dict(sorted(priorities.items(),reverse=True, key=lambda x: x[1]))
    return ordered_prior_route

def _priority_connection(priority_route:dict, cost_matrix, mean_routes, std_routes, mean_weight, std_weigth, distance_weight) -> dict:
    """
Give us the priority connections dictionary for each route

Parameters
-----------

priority_route: dict
    The dictionary with the priority routes

cost_matrix: dict of dicts
    it's a dict of dicts with the points to visit as a keys and
    the value is another dictionary with the points to visit and
    the value of going for the first point to the second. It's a
    dictionary representation of a matrix.

mean_routes: dict
    A dict with the mean distance/cost of all points one each other

std_routes: dict
    A dict with the standar deviation distance/cost of all points one each other

mean_weight: float
    The ponderation of the mean for all points, higher number make points
    with higher mean (more distance from others) have large values and priorice
    their connections

std_weight: float
    The ponderation of the standar deviation for all points, higher number
    make points with higher deviation (more diferents of distance between points)
    have larger values and priorice their connections

distance_weight: float
    The ponderation of the distance between all points, higher distance_weight 
    make points with higher distances between them have large values

Return
-------
A dict of dicts every point connections ordered by priority
    """



    dict_prior_connections_ordered = {}
    base_order = priority_route.keys()
    for id_route in base_order:
        prior_connections = {}
        for given_route in base_order:
            if id_route != given_route:
                prior_connections[given_route] = ((mean_routes[given_route]**mean_weight) * (std_routes[given_route]**std_weigth)) /  cost_matrix[given_route][id_route] ** distance_weight
        dict_prior_connections_ordered[id_route] = dict(sorted(prior_connections.items(), reverse=True, key=lambda x: x[1]))
    return dict_prior_connections_ordered

def _compute_mean_std_dicts(random_route, cost_matrix, lenght_route):
    """
Compute the mean and the standar deviation distance between all points
one each other

Parameters
----------
random_route: list
    A list of ints with all our points

cost_matrix: dict of dicts
    it's a dict of dicts with the points to visit as a keys and
    the value is another dictionary with the points to visit and
    the value of going for the first point to the second. It's a
    dictionary representation of a matrix.

length_route: int
    The amount of points

Return
-------
A tuple with two dicts and the mean and standar deviation of all points
    """
    mean_routes = {}
    std_routes = {}
    for id_route in random_route:
        values = list(cost_matrix[id_route].values())
        mean = sum(values) / lenght_route
        mean_routes[id_route] = mean
        std_routes[id_route] = _std(values, mean, lenght_route)
    return mean_routes, std_routes


def _check_route(node_dictionary, length_route):
    """
Check if the connection between two points create a circular route before the algorithm
finish

Parameters
-----------
node_dictionary dict
    A dictionary with the key as integer representing the node point and a value
    as a Node object

length_route int
    The amount of points


Returns
-----------
 There is two possibles returns
    * A boolean if the route its possible or not
    * if it's a circular route, but we have cover all points (length), returns the
    route path.
    """

    for k in node_dictionary.keys():
        if node_dictionary[k].neighbours:
            path = [node_dictionary[k].node, node_dictionary[k].neighbours[0]]
            counter = 0
            while counter < length_route-1:
                last_route = path[-1]
                next_route = node_dictionary[last_route]
                next_route_neighbours = next_route.neighbours
                #Comprobamos si están los dos en el path, si están hay una ruta circular, por lo que retornamos false
                if len(next_route_neighbours) == 2:
                    if next_route_neighbours[0] in path and next_route_neighbours[1] in path:
                        if len(path) != length_route:
                            return False
                        else:
                            return path
                    elif next_route_neighbours[0] in path:
                        path.append(next_route_neighbours[1])
                        counter += 1
                    elif next_route_neighbours[1] in path:
                        path.append(next_route_neighbours[0])
                        counter +=1
                else: #SI un nodo no tiene dos rutas, entonces no puede ser circular  por loq ue pasamos
                    counter +=1
    return True



def _define_true_route(ids_route):
    """
Only for developt Purposes, a true good route.
    """
    true_dict = {i:Node(i) for i in ids_route}
    true_dict[10000007].neighbours = 46
    true_dict[10000007].neighbours = 47
    true_dict[47].neighbours = 30
    true_dict[47].neighbours = 10000007
    true_dict[30].neighbours = 23
    true_dict[30].neighbours = 47
    true_dict[23].neighbours = 59
    true_dict[23].neighbours = 30
    true_dict[59].neighbours = 4
    true_dict[59].neighbours = 23
    true_dict[4].neighbours = 91
    true_dict[4].neighbours = 59
    true_dict[91].neighbours = 25
    true_dict[91].neighbours = 4
    true_dict[25].neighbours = 1
    true_dict[25].neighbours = 91
    true_dict[1].neighbours = 31
    true_dict[1].neighbours = 25
    true_dict[31].neighbours = 5
    true_dict[31].neighbours = 1
    true_dict[5].neighbours = 12
    true_dict[5].neighbours = 31
    true_dict[12].neighbours = 7
    true_dict[12].neighbours = 5
    true_dict[7].neighbours = 26
    true_dict[7].neighbours = 12
    true_dict[26].neighbours = 27
    true_dict[26].neighbours = 7
    true_dict[27].neighbours = 65
    true_dict[27].neighbours = 26
    true_dict[65].neighbours = 44
    true_dict[65].neighbours = 27
    true_dict[44].neighbours = 32
    true_dict[44].neighbours = 65
    true_dict[32].neighbours = 46
    true_dict[32].neighbours = 44
    true_dict[46].neighbours = 10000007
    true_dict[46].neighbours = 32
    return true_dict


class NodeValueError(Exception):
    """
NodeValueError Exception is throw when a node have two neighbours and is added 
one more
    """
    def __init__(self, node, neighbours):
        self.node = node
        self.neigbours = neighbours
    def __str__(self):
        return f"A node only can have two neighbours. The node {self.node} has {self.neigbours} neighbours"



class Node:
    """
Node Class

This Class create a Node, a node is a point that can have a maximum of two connections

Parameters
-----------

node int
    The point that will be connected with two other points

--------------------------------------------------------------------------------------

Attributes
-----------
#Dataset - id of origin - id of destiny - column to transform (in this case the hour)
neighbours list
    neigbours is a list of int, each time you add a neighbours you to the list is a
    connection with the node. Remember that a node only can have a maximum of two
    neighbours


--------------------------------------------------------------------------------------

Special Methods
---------------

__repr__ str
    The Node and their neighbours

__len__ int
    Number of neighbours

--------------------------------------------------------------------------------------

Example
--------

node1 = Node(1)

#Add neighbours
node1.neighbours = 4
node1.neighbours = 7

#Print a node
print(node1)

# length of node
print(len(node1))
    """

    def __init__(self, node):
        self.node = node
        self._neighbours = []

    def __repr__(self):
        return str((self.node, self.neighbours))

    def __len__(self):
        return len(self._neighbours)

    @property
    def neighbours(self):
        return self._neighbours

    @neighbours.setter
    def neighbours(self, x):
        if len(self._neighbours) < 2:
            self._neighbours.append(x)
        else:
            raise NodeValueError(self.node, self.neighbours)
        




from ._base_simulated_annealing._base_simulated_annealing import SimulatedAnnealing
from ._base_ensemble_simulated_annealing import EnsembleSimulatedAnnealing


__all__ = ['SimmulatedAnnealing', "EnsembleSimulatedAnnealing"]




#cython: language_level = 3, boundscheck = False, wraparound = False, nonecheck = False, embedsignature=True
from ._base_genetics._base_genetic import Genetic
from ..._validators._validators import _intenger_validator
from multiprocessing import cpu_count, Pool

class EnsembleGenetic(Genetic):
    """
Instantiate the EnsembleGenetic class that inherit from Genetic class

Parameters:
-----------
n_genetics: int, default=10
    The  number of Genetics algorithm that will be thrown, more
    algorithms you will achieve a better result, but will be more
    slower

p_c: float32, default=0.6
    between 0 and 1 the probability of crossover. In each generation
    at start the algorithm generate two random probabilities these
    random probabilies are the choosen when crossover will take
    place. If the random number is 4 that mean that the number in
    position and the following numbers will be now at start. And the
    first four numbers will be at the end.

p_m : float32, default=0.4
    it's mutate, random probabilies are choosen when it will take
    place. If the random numbers are 2 and 8 for example, the 
    numbers located at that index will swap positions.

pop: int32, default=400
    The number of population (different random solutions), more
    population, more probabilities to find different solutions but 
    make the algorithm more time expensive. If the number is too
    low, the solution will be bad, because the algorithm can't 
    iterate over the loss function with the enough amount of data

gen: int32 Default=1000
    The number of generations. More generations make the algorithm
    achieve better solution but also make the algorithm take more
    time. This can be solved with early_stoping parameter to make
    the algorithm stop when is not improving. If the number of
    generation is low the algorithm may not converge

k: int32, default=3
    The number of individual populations (solutions) that
    will fight for being the best. When we are inside of a
    generation the algorithm take a subsample of numbers k. The best
    k (individual) solution is choosen as a parten for this generation.

early_stopping: int32, default=None
    If it's None, the algorithm will finish when the last generation  
    is performed. If it's other unsigned integer (positive integer
    number) if the algorithm can't improve the result while X
    generations (where x is early_stopping number) the algorithm
    will stop.

max_time_work: float32, default=8
    the number of ours that a employ can work per day. For example
    if it's 8 hours, the algorithm will force that a route have to
    finish after the 8 hours have been completed, making the
    employeed come back home. it's a time constraint.

extra_cost: float32, default=0
    if it's 0 anything happend. If it's >0 in combination with
    max_time_work when the max_time_work is reached, extra_cost is
    applied. This add a extra cost to the solution each time that
    max_time_work is reached. It's like extra pay for the worker 
    each time max_time_work is completed (journey).

people: int32, default=1
    The number of people that you use in each route, for example if
    you need two truck drivers. that's another contstraint. That
    will multiply the time_costs and the extra_cost. Not the travel
    cost because it's assumed that both go in the same vehicle.

n_jobs: int default=0
    The number of jobs to work in parallel the more workers, the
    faster will go. Use this parameter with caution, maybe can 
    collapse the computer if you select a lot jobs. 


--------------------------------------------------------------------------

Methods:
--------

fit(self, list route_example, dict time_matrix, dict fuel_matrix)
    Execute the algorithm and give back the best solution

    Parameters
    ----------
    route_example: list
        it's a list that contain a random example of ints ids routes. But it's
        mandatory that the origin place, be the first id of the list

    time_costs: dict
        it's a dict of dicts with the points to visit as a keys and
        the value is another dictionary with the points to visit and
        the value of going for the first point to the second. It's a
        dictionary representation of a matrix. For example if we
        have 3 ID 1, 2 and 3 the dict will be like this:

            {
            1:{
                1:0,
                2:x,
                3:y
                },
            2:{
                1:x,
                2:0,
                3:z
                },
            3:{
                1:y,
                2:z,
                3:0
                }
            }

        This is just a (3, 3) symmetric matrix with the cost in
        time from one point to another with column and index.

            1   2   3
        1   0   x   y
        2   x   0   z
        3   y   z   0
    
    fuel_costs: dict
        For a extended explanation go up to time_matrix. it's
        the fuel cost to go from one point to another.

    Returns:
    --------
    A very optimized route that solve the MTSPTW problem.

---------------------------------------------------------------------------


Example
--------
from sklearn_route.datasets import load_barcelona
from sklearn_route.preprocessing import matrix_to_dict
from sklearn_route.metaheuristics.genetics import EnsembleGenetic

df_barcelona = load_barcelona()

#Dataset - id of origin - id of destiny - column to transform (in this case the hour)
time_matrix = matrix_to_dict(df_barcelona, "id_origin", "id_destinity", "hora")

#Dataset - id of origin - id of destiny - column to transform (in this case the cost)
cost_matrix = matrix_to_dict(df_barcelona, "id_origin", "id_destinity", "cost")

#Create a random route, it's will needed to initiate the algorithm
route_example = list(dict.fromkeys(cost_matrix_df).keys())

#Instantiate the algorithm
eg = EnsembleGenetic(n_genetics=20, n_jobs=4, p_m = 0.6, pop=400, gen=2000, max_time_work=6, extra_cost=12.83)

#random route - time_matrix - cost_matrix
result = eg.fit(route_example, time_matrix, cost_matrix)

#Printing the best route
print(result)
    """
    
    def __init__(self, n_genetics:int= 10,  p_c:float=0.6, p_m:float=0.4, pop:int=400, gen:int=1000, k:int=3, early_stopping=None,
                  max_time_work:float=8., extra_cost:float=10., people:int=1, n_jobs:int=1):
        super().__init__(p_c, p_m, pop, gen, k, early_stopping, max_time_work, extra_cost, people)
        self.n_genetics = _intenger_validator(n_genetics, "n_genetics")
        self.n_jobs = _intenger_validator(n_jobs, "n_jobs")

    def fit(self, route_example:list, time_matrix:dict, cost_matrix:dict) -> list:
        if self.n_jobs > 1:
            pool = Pool(processes=self.n_jobs)
            list_to_pool = []
            for i in range(self.n_genetics):
                list_to_pool.append([Genetic(self.p_c, self.p_m, self.pop, self.gen, self.k, self.early_stopping,
                                    self.max_time_work, self.extra_cost, self.people),
                 route_example, time_matrix, cost_matrix])
            list_results = pool.starmap(Genetic.fit, list_to_pool)
            min_value = list_results[0][0]
            result = list_results[0]
            for i in range(self.n_genetics):
                if list_results[i][0] < min_value:
                    result = list_results[i]
            return result
        else:
            result = []
            value = []
            for i in range(self.n_genetics):
                algo_result = super().fit(route_example, time_matrix, cost_matrix)
                result.append(algo_result)
                value.append(algo_result[0])
            return result[value.index(min(value))]




import pandas as pd
import os
_path = os.path.join(os.path.dirname(__file__))

__all__ = ["load_alicante_murcia", "load_barcelona", "load_madrid", "load_valencia"]
#TODO: describe functions




def load_alicante_murcia():
    return {
        "DataFrame":pd.read_pickle(_path + "/Alicante-Murcia_places.pkl"),
        "DESCR": _read_txt(_path + "/DESCR.txt"),
        "feature_names": _read_txt(_path + "/columns.txt")
    }

def load_barcelona():
    return {
        "DataFrame":pd.read_pickle(_path + "/Barcelona_places.pkl"),
        "DESCR": _read_txt(_path + "/DESCR.txt"),
        "feature_names": _read_txt(_path +"/columns.txt")
    }

def load_madrid():
    return {
        "DataFrame":pd.read_pickle(_path + "/Madrid_places.pkl"),
        "DESCR": _read_txt(_path + "/DESCR.txt"),
        "feature_names": _read_txt(_path + "/columns.txt")
    }

def load_valencia():
    return {
        "DataFrame":pd.read_pickle(_path + "/Valencia.pkl"),
        "DESCR": _read_txt(_path + "/DESCR.txt"),
        "feature_names": _read_txt(_path + "/columns.txt")
    }        


def load_argentina(mode="big"):
    return _loader("/ar9152.tsp", mode)

def load_burma(mode="big"):
    return _loader("/bm33708.tsp", mode)

def load_china(mode="big"):
    return _loader("/ch71009.tsp", mode)

def load_canada(mode="big"):
    return _loader("/ca4663.tsp", mode)

def load_djibouti(mode="big"):
    return _loader("/dj38.tsp", mode)

def load_egypt(mode="big"):
    return _loader("/eg7146.tsp", mode)

def load_ireland(mode="big"):
    return _loader("/ei8246.tsp", mode)

def load_finland(mode="big"):
    return _loader("/fi10639.tsp", mode)

def load_greece(mode="big"):
    return _loader("/gr9882.tsp", mode)

def load_honduras(mode="big"):
    return _loader("/ho14473.tsp", mode)

def load_italy(mode="big"):
    return _loader("/it16862.tsp", mode)

def load_japan(mode="big"):
    return _loader("/ja9847.tsp", mode)

def load_kazakhstan(mode="big"):
    return _loader("/kz9976.tsp", mode)

def load_luxembourg(mode="big"):
    return _loader("/lu980.tsp", mode)

def load_morocco(mode="big"):
    return _loader("/mo14185.tsp", mode)

def load_oman(mode="big"):
    return _loader("/mu1979.tsp", mode)

def load_nicaragua(mode="big"):
    return _loader("/nu3496.tsp", mode)

def load_panama(mode="big"):
    return _loader("/pm8079.tsp", mode)

def load_qatar(mode="big"):
    return _loader("/qa194.tsp", mode)

def load_rwanda(mode="big"):
    return _loader("/rw1621.tsp", mode)

def load_sweden(mode="big"):
    return _loader("/sw24978.tsp", mode)

def load_tanzania(mode="big"):
    return _loader("/tz6117.tsp", mode)

def load_uruguay(mode="big"):
    return _loader("/uy734.tsp", mode)

def load_vietnam(mode="big"):
    return _loader("/vm22775.tsp", mode)

def load_sahara(mode="big"):
    return _loader("/wi29.tsp", mode)

def load_yemen(mode="big"):
    return _loader("/ym7663.tsp", mode)

def load_zimbabwe(mode="big"):
    return _loader("/zi929.tsp", mode)



def _loader(url, mode="big"):
    dictionary_dataset = {
        "DESCR": _read_txt(_path + "/DESCR.txt"),
        "feature_names": _read_txt(_path + "/columns.txt")
    }
    if mode == "big":
        dictionary_dataset["DataFrame"] = _read_tsp(_path + url)
        return dictionary_dataset
    elif mode == "medium":
        df =_read_tsp(_path + url)
        dictionary_dataset["DataFrame"] = df.sample(int(len(df) *0.2), random_state=2019)
        return dictionary_dataset
    elif mode== "small":
        df =_read_tsp(_path + url)
        dictionary_dataset["DataFrame"] = df.sample(int(len(df) *0.005), random_state=2019)
        return dictionary_dataset
    else:
        raise ValueError(f"The mode '{mode}' is not availabre, the modes availables are: big, medium, small")




def _read_txt(path:str) -> list:
    """sumary_line
    
    Keyword arguments:
    argument -- description
    Return: return_description
    """
    
    with open(path, "r") as file:
        readed_file = file.read()
    return readed_file


def _read_tsp(url):
    return pd.read_csv(url, skiprows=7, skipfooter=1, names=["id", "latitude", "longitude"], sep="\s")



from sklearn.cluster import KMeans, AffinityPropagation
from sklearn.mixture import GaussianMixture, 
from kmeans_truncate import KMeansTruncate

__all__ = ["KMeans", "DBSCAN", "KMeansTruncate", "AffinityPropagation", "GaussianMixture"]



import numpy as np
import pandas as pd
from itertools import combinations
# import plotly.express as px
# import plotly.offline as pyo
from googlemaps import Client
from time import sleep
import pickle
import warnings
from tqdm import tqdm

def dfcolumn_to_dict(dataframe, id_origin, id_destinity, compute_column):
    """
Takes a DataFrame and return a dict of dicts that represent a 2D matrix with costs

Parameters
-----------
dataframe: DataFrame
    The dataframe

id_origin: str
    The column of the DataFrame with the id of origin

id_destinity: str
    The column of the DataFrame with the id of destinity

compute_columns: str
    The columns with the values to compute time/distance
    """

    columns = dataframe[id_origin].drop_duplicates().values
    cluster_len = len(dataframe[id_destinity].drop_duplicates())
    empty_matrix = np.zeros([cluster_len, cluster_len])
    start = 0
    count = 0
    finish = cluster_len
    for n in range(1, cluster_len):
        count += 1
        empty_matrix[n-1][n-1:cluster_len] = dataframe[compute_column][start:finish]
        start = finish
        finish = start + cluster_len - count
    clusters_cost_matrix = pd.DataFrame(empty_matrix + empty_matrix.T, columns=columns, index=columns).to_dict()

    return clusters_cost_matrix

def normalize(df_nodes, lat, lon):
    """
Take a Dataframe and normalize the latitude and the longitude columns, works very well with
SOM algorithm.

Parameters
-----------
df_nodes: dataframe
    The dataframe with the latitude and longitude columns, can have more columns.

lat: string
    The name of the latitude column

lon: string
    The name of the longitude column

Return
-------
The columns latitude and longitude normalize
    """
    ratio = (df_nodes[lat].max() - df_nodes[lat].min()) / (df_nodes[lon].max() - df_nodes[lon].min()), 1
    ratio = np.array(ratio) / max(ratio)
    norm = df_nodes[[lat, lon]].apply(lambda c: (c - c.min()) / (c.max() - c.min()))
    df_nodes[[lat, lon]] = norm.apply(lambda p: ratio * p, axis=1)
    return df_nodes

def df_to_tuple(dataframe):
    """
Parse a DataFrame into node of tuples
    """
    nodes = []
    for n in dataframe.to_numpy():
        n = list(n)
        n[0] = int(n[0])
        nodes.append(tuple(n))

    return tuple(nodes)



# def _swap_point(string):
#     string = str(string).replace(".", "")
#     string = string[:2] + "." + string[2:]
#     return float(string)

# def _plot_points(df, lat, lon, save=False, path=None):
#     fig = px.scatter_mapbox(df, lat='latitude', lon='longitude')
#     fig.update_layout(mapbox_style="open-street-map")
#     fig.update_layout(margin={"r":0,"t":0,"l":0,"b":0})
#     fig.show(filename=path)
#     if save:
#         pyo.plot(fig, filename=path, auto_open=False)



class DataLossWarning(UserWarning):
    def __str__(self):
            return " Warning DataLossWarning: The data has None values, that have been deleted. Please check your data if you don't want to loss nodes"


class CostScraper:  
    """
CostScraper uses the Google API to extract the distance between points.

YOU NEED A GOOGLE ACCOUNT, AND YOU HAVE TO PAY GOOGLE FOR EVERY EXTRACTION. Please
visit https://developers.google.com/maps/documentation/distance-matrix/usage-and-billing
for more info.

Scikit Route is a completly free open software package, we don't have any economic
interest on this. It's just to help people to have and easy way to get the distance in time
and in meters. Scikit Route don't encourage or promote the use of the Google API.

Parameters:
-----------

p_c: float32, default=0.6
    between 0 and 1 the probability of crossover. In each generation
    at start the algorithm generate two random probabilities these
    random probabilies are the choosen when crossover will take
    place. If the random number is 4 that mean that the number in
    position and the following numbers will be now at start. And the
    first four numbers will be at the end.

api: str
    The API credential to the distance matrix Google service

nodes: list
    A list of lists with the ID, latitude and longitude of every node. CostScraper takes 
    all the info and combine every node with the rest, using symmetric routes
    (same distance go and back). This is done with itertools.combinations.
    The most important part is respect the order of the tuple. This is an input example:

    nodes = [
        [1, 26.0336111, 51.2002778],
        [2, 26.0488889, 51.0569444],
        [3, 26.05, 51.25],
        [4, 26.0502778, 51.2975],
        [5, 26.0505556, 51.1358333],
    ]

mode: str, {"driving","walking","bicycling","transit"} default="driving"
    The mode of transport that Google API Distance Matrix offer.

------------------------------------------------------------------------------------------------------------

Method:
---------

scrap():
    With the credentials, the nodes and the mode, scrap all distances from all nodes to
    others and return a tuple of connections between all nodes and the time and meters
    needed. Example:
    
    nodes_connected = [
    (1,
    25.01,
    51.0394444,
    'Al Karaana Road، Al Kir anah, Qatar',
    2,
    25.2847222,
    51.5552778,
    'Ras Bu Abboud Street، Doha, Qatar',
    71104,
    3712),
    (1,
    25.01,
    51.0394444,
    'Al Karaana Road، Al Kir anah, Qatar',
    3,
    25.2861111,
    51.5041667,
    'Barwa Alsadd, C Ring Rd, Doha, Qatar',
    66279,
    3287),
    ...
    ]

    The elements of each tuple are:
    - First node id
    - Latitude of first node
    - Longitude of first node
    - Address of first node
    - Second node id (The connection will be first node with second)
    - Latitude of seconde node
    - Longitude of sencond node
    - Address of sencond node
    - Distance in seconds from first node to second node and viceversa (symmetric)
    - Distance in meters  from first node to second node and viceversa (symmetric)

    ¡Caveat!
    The None values are distances not availables in the Google API.  So sometimes
    it's needed to continue processing the result of this method by hand.

to_pickle(filename):
    Parameter
    ---------
    filename: The path with the .pkl extension to save the results scraped, strongly 
    recommended to use, in order to not to lose the information.

pandas(): 
    The data are parse to a Pandas.DataFrame
    """

    def __init__(self, api, nodes, mode="driving"):
        self._api = Client(api)
        self._nodes = combinations(nodes, r=2)
        self.mode = mode
        self._ids = [n[0] for n in nodes]

    @property
    def api(self):
        return self._api.key

    @api.setter
    def api(self, credential):
        self._api = Client(credential)

    @api.deleter
    def apit(self):
        del self._api

    @property
    def nodes(self):
        return self._nodes
    
    @nodes.setter
    def nodes(self, values):
        self._nodes = combinations(values, r=2)
        self._ids = [n[0] for n in values]
    
    @nodes.deleter
    def nodes(self):
        del self._nodes

    def scrap(self):
        """
        Al ejecutar este método se envia la petición a Google y este nos devuelve las distancias,
        en metros y segundos. **CUIDADO EJECUTAR ESTE MÉTODO TIENE UN COSTE MONETARIO**
        """        

        scraped_conections = []
        for connection in tqdm(self._nodes):
            sleep(0.01)
            res_dict = self._api.distance_matrix(tuple(connection[0][1:]), tuple(connection[1][1:]), mode=self.mode)
            
            elements = res_dict["rows"][0]["elements"][0]
            if res_dict["status"] == "OK" and elements["status"] == "OK":
                scraped_conections.append((*connection[0], res_dict["origin_addresses"][0], *connection[1],
                                            res_dict["destination_addresses"][0], elements["distance"]["value"],
                                            elements["duration"]["value"]))
            else:
                scraped_conections.append(tuple([None]*10))
        
        self.scraped_conections = scraped_conections
        return scraped_conections
    
    def to_pickle(self, filename):
        if ".pkl" == filename[-4:]:
            raise TypeError(f'filename must finish in ".pkl" this is the pickle extension and is finishing {filename[:-4]}')
        with open(filename, "wb") as f:
            pickle.dump(self.scraped_conections, f)

    def pandas(self):
        columns = ["id_origin", "lat_origin", "lon_origin", "address_origin", "id_destinity", 
                  "lat_destinity", "lon_destinity", "address_destinity", "meters", "seconds"]
        return pd.DataFrame(self.scraped_conections, columns=columns)



def matrix_parse(connections, mode="meters"):
    """
Parse a list of tuples connected with the format of the CostScraper output into a matrix.
Which is a list of tuples each tuple representing a connection between two nodes

Parameters:
------------

connections: list
    The list of tuples connection  with the format of the CostScraper output into a matrix.
    Which is a list of tuples each tuple representing a connection between two nodes. Example:

     nodes_connected = [
    (1,
    25.01,
    51.0394444,
    'Al Karaana Road، Al Kir anah, Qatar',
    2,
    25.2847222,
    51.5552778,
    'Ras Bu Abboud Street، Doha, Qatar',
    71104,
    3712),
    (1,
    25.01,
    51.0394444,
    'Al Karaana Road، Al Kir anah, Qatar',
    3,
    25.2861111,
    51.5041667,
    'Barwa Alsadd, C Ring Rd, Doha, Qatar',
    66279,
    3287),
    ...
    ]

    The elements of each tuple are:
    - First node id
    - Latitude of first node
    - Longitude of first node
    - Address of first node
    - Second node id (The connection will be first node with second)
    - Latitude of seconde node
    - Longitude of sencond node
    - Address of sencond node
    - Distance in meters from first node to second node and viceversa (symmetric)
    - Distance in seconds  from first node to second node and viceversa (symmetric)

mode: str {"meters", "time"} default="meters"
    If meters, the idx -2 of each tuple is selected else, the idx -1 is selected.
    """
    ids = []
    check_none = False
    for con in connections:
        if not check_none and None in con:
            check_none = True
        ids.append(con[0])
    
    if check_none:
        warnings.warn(str(DataLossWarning), DataLossWarning)

    length_connections = len(list(filter(lambda x: x[0] == ids[0], connections))) +1
    ids = dict.fromkeys([con[0] for con in connections if con[0] is not None])

    matrix = np.empty((length_connections, length_connections))

    if mode == "seconds":
        idx = -1
    elif mode == "meters":
        idx = -2
    else:
        raise ValueError('Please select an available mode. The modes availables are "meters" and "seconds"')

    counter = 0
    for node_id in ids:
        selected = filter(lambda x: x[0] == node_id, connections)
        matrix[counter, counter+1:] = [sel[idx] for sel in selected]
        counter +=1

    return matrix

def matrix_to_dict(dict_of_dicts):
    """
2D array with cost of each node with the rest, return a dict of dicts ready for 
the algorithm

Parameters
-----------
dict_of_dicts: 2d array
    the matrix that will be converted into a dict of dicts.

Return
-------
dict of dicts with the connections between nodes.

    """
    clean_dict = {}
    for key in dict_of_dicts:
        temp_dict = {}
        for k, v in dict_of_dicts[key].items():
            if v !=0:
                temp_dict[k] = v
        clean_dict[key] = temp_dict



import os.path
from setuptools import setup, find_packages
from setuptools.extension import Extension
from Cython.Build import cythonize
import Cython.Compiler.Options
Cython.Compiler.Options.annotate = True
import re

extensions = cythonize([
    Extension(
        "skroute.heuristics.brute._base_brute_force._base_brute_force",
        sources=["skroute/heuristics/brute/_base_brute_force/_base_brute_force.pyx"]),
    Extension(
        "skroute.metaheuristics.genetics._base_genetics._utils_genetic",
        sources=["skroute/metaheuristics/genetics/_base_genetics/_utils_genetic.pyx"]),
    Extension(
        "skroute._utils._utils",
        sources=["skroute/_utils/_utils.pyx"]),
    Extension(
        "skroute.metaheuristics.genetics._base_genetics._base_genetic",
        sources=["skroute/metaheuristics/genetics/_base_genetics/_base_genetic.pyx"]),
    Extension(
        "skroute.metaheuristics.simulated_annealing._base_simulated_annealing._base_simulated_annealing",
        sources=["skroute/metaheuristics/simulated_annealing/_base_simulated_annealing/_base_simulated_annealing.pyx"]),
    Extension(
        "skroute.metaheuristics.simulated_annealing._base_simulated_annealing._utils_sa",
        sources=["skroute/metaheuristics/simulated_annealing/_base_simulated_annealing/_utils_sa.pyx"]),
    ])



HERE = os.path.abspath(os.path.dirname(__file__))


with open(os.path.join(HERE, "README.md")) as fid:
    README = fid.read()

with open(os.path.join(HERE, "requirements.txt")) as f:
    REQUIREMENTS = f.read().splitlines()


setup(
    name="skroute", 
    version="1.0.0a1 ", 
    description="Compute Routes easy and fast",
    long_description=README, 
    long_description_content_type="text/markdown",  
    url="https://github.com/arubiales/scikit-route", 
    author="Alberto Rubiales", 
    author_email="al.rubiales.b@gmail.com", 
    license="MIT License", 
    classifiers=[ 
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9"
    ],
    packages=find_packages(),
    setup_requires=[
        'setuptools>=18.0',
        "cython"
    ],
    install_requires=REQUIREMENTS,
    ext_modules = extensions,
    include_package_data=True,
    package_data={"": ["datasets/*.csv", "datasets/*.tsp", "datasets/*.txt", "datasets/*.pkl"]},
)



from itertools import permutations
import os, sys
sys.path.append("/".join(os.path.abspath("").split("/")[:-1]))
from ..._validators import _intenger_validator
from ..._utils._utils import _cost, _final_cost

#NOTE: Pasar a Cython

class BruteForce:

    def __init__(self, max_time_work=8, extra_cost=0, people=1):
        self.max_time_work = max_time_work
        self.extra_cost = extra_cost
        self.people = people

    def fit(self, route_example, time_matrix, cost_matrix):
        length = len(route_example)
        route_to_permute = route_example[1:]
        home = route_example[0]
        min_cost = _cost(route_example, length, time_matrix, cost_matrix, self.max_time_work, self.extra_cost, self.people)
        min_route = route_example
        for route in permutations(route_to_permute):
            list_route = list(route)
            list_route.insert(0, home)
            route_cost = _cost(list_route, length, time_matrix, cost_matrix, self.max_time_work, self.extra_cost, self.people)
            if route_cost < min_cost:
                min_route = list_route
                min_cost  = route_cost

        return min_cost, _final_cost(min_route, length, time_matrix, self.max_time_work)

